/*
 *  SIPdump | Martin J. Muench | mjm@codito.de | 2006
 *
 *  SIP login sniffer
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pcap.h>

#include "global.h"
#include "wrap.h"

#define VERSION     "0.1"

int opterr = 0; /* shutup getopt() */

#define SIP_LINE_LEN 1024 /* Maximum length of sip protocol lines */
#define MAX_SIP_CON  128  /* Maximum parallel SIP connections     */

typedef struct
{
  int       active;
  uint32_t  client_ip;
  uint16_t  client_port;
  uint32_t  server_ip;
  uint16_t  server_port;
  char      buffer[SIP_LINE_LEN];
} sip_conn_t;

typedef struct
{
  uint32_t  client_ip;
  uint16_t  client_port;
  uint32_t  server_ip;
  uint16_t  server_port;
} conn_t;

/* Function declarations */
static void sniff_logins(unsigned char *, const struct pcap_pkthdr *, const unsigned char *);
static void parse_payload(const conn_t *, unsigned char *, size_t);
static int  parse_sip_proto(char *, size_t, unsigned char *, size_t);
static int  find_sip_client_proto(unsigned char *, size_t);
static void manual_mode();
static void usage(const char *);
static void parse_n_write_login_data(uint32_t, uint32_t, char *);

static char        *logins_file=NULL; /* dump file             */
static unsigned int num_logins=0;     /* sniffed login counter */

#define SNAP_LEN        1518
#define SIZE_ETHERNET   14
#define ETHER_ADDR_LEN	6

/* Ethernet header */
struct ethernet_header {
  unsigned char  ether_dhost[ETHER_ADDR_LEN];
  unsigned char  ether_shost[ETHER_ADDR_LEN];
  unsigned short ether_type;
};

/* IP header */
struct ip_header {
  unsigned char  ip_vhl;
  unsigned char  ip_tos;
  unsigned short ip_len;
  unsigned short ip_id;
  unsigned short ip_off;
#define IP_RF 0x8000            /* reserved fragment flag */
#define IP_DF 0x4000            /* dont fragment flag */
#define IP_MF 0x2000            /* more fragments flag */
#define IP_OFFMASK 0x1fff       /* mask for fragmenting bits */
  unsigned char  ip_ttl;
  unsigned char  ip_p;
  unsigned short ip_sum;
  struct  in_addr ip_src,ip_dst;
};

#define IP_HL(ip)               (((ip)->ip_vhl) & 0x0f)
#define IP_V(ip)                (((ip)->ip_vhl) >> 4)

/* TCP header */
typedef unsigned int tcp_seq;

struct tcp_header {
  unsigned short th_sport;
  unsigned short th_dport;
  tcp_seq th_seq;
  tcp_seq th_ack;
  unsigned char  th_offx2;
#define TH_OFF(th)      (((th)->th_offx2 & 0xf0) >> 4)
  unsigned char  th_flags;
#define TH_FIN  0x01
#define TH_SYN  0x02
#define TH_RST  0x04
#define TH_PUSH 0x08
#define TH_ACK  0x10
#define TH_URG  0x20
#define TH_ECE  0x40
#define TH_CWR  0x80
#define TH_FLAGS        (TH_FIN|TH_SYN|TH_RST|TH_ACK|TH_URG|TH_ECE|TH_CWR)
  unsigned short th_win;
  unsigned short th_sum;
  unsigned short th_urp;
};

struct udp_header {
  u_int16_t       uh_sport;
  u_int16_t       uh_dport;
  u_int16_t       uh_ulen;
  u_int16_t       uh_sum;
};


int main(int argc, char *argv[])
{
  char *dev=NULL, *tcpdump_file=NULL, errbuf[PCAP_ERRBUF_SIZE], filter_exp[] = "tcp or udp";
  pcap_t *handle;
  struct bpf_program fp;
  bpf_u_int32 mask, net;
  int c, manual=0;

  printf("\nSIPdump %s  ( MaJoMu | www.remote-exploit.org ) \n"  \
         "------------------------------------------------\n\n", VERSION);

  /* Get params */
  while((c = getopt(argc, argv, "d:i:mf:")) != -1) {
    switch(c) {
    case 'd':
      logins_file = (char *)Malloc(strlen(optarg)+1);
      strcpy(logins_file, optarg);
      break;
    case 'i':
      dev = (char *)Malloc(strlen(optarg)+1);
      strcpy(dev, optarg);
      break;
    case 'm':
      manual = 1;
      break;
    case 'f':
      tcpdump_file = (char *)Malloc(strlen(optarg)+1);
      strcpy(tcpdump_file, optarg);
      break;
    default:
      usage("Invalid arguments");
    }
  }
  if(logins_file == NULL)
    usage("Invalid arguments");

  if(manual) {
    manual_mode();
    exit(EXIT_SUCCESS);
  }

  if(geteuid() != 0 && tcpdump_file == NULL) {
    printf("* You need to have root privileges to run this program\n");
    exit(EXIT_SUCCESS);
  }

  if(tcpdump_file != NULL) {
    printf("* Using tcpdump data file '%s' for sniffing\n", tcpdump_file);

    handle = pcap_open_offline(tcpdump_file, errbuf);
    if (handle == NULL) {
      fprintf(stderr, "* Cannot open device %s: %s\n", dev, errbuf);
      exit(EXIT_FAILURE);
    }
  }
  else {
    /* Get interface if not specified on command line */
    if(dev == NULL) {
      dev = pcap_lookupdev(errbuf);
      if (dev == NULL) {
	fprintf(stderr, "* Couldn't find default device: %s\n", errbuf);
	exit(EXIT_FAILURE);
      }
    }

    printf("* Using dev '%s' for sniffing\n", dev);
    
    /* get network number and mask associated with capture device */
    if (pcap_lookupnet(dev, &net, &mask, errbuf) == -1) {
      fprintf(stderr, "* Couldn't get netmask for device %s: %s\n", dev, errbuf);
      net = 0;
      mask = 0;
    }
  
    /* open capture device */
    handle = pcap_open_live(dev, SNAP_LEN, 1, 1000, errbuf);
    if (handle == NULL) {
      fprintf(stderr, "* Cannot open device %s: %s\n", dev, errbuf);
      exit(EXIT_FAILURE);
    }

    /* make sure we're capturing on an Ethernet device [2] */
    if (pcap_datalink(handle) != DLT_EN10MB) {
      fprintf(stderr, "* Invalid device '%s' (non ethernet)\n", dev);
      exit(EXIT_FAILURE);
    }
  }

  /* compile the filter expression */
  if (pcap_compile(handle, &fp, filter_exp, 0, net) == -1) {
    fprintf(stderr, "* Invalid packet filter %s\n", pcap_geterr(handle));
    exit(EXIT_FAILURE);
  }

  /* apply the compiled filter */
  if (pcap_setfilter(handle, &fp) == -1) {
    fprintf(stderr, "* Installing filter failed: %s\n", pcap_geterr(handle));
    exit(EXIT_FAILURE);
  }

  printf("* Starting to sniff with filter '%s'\n\n", filter_exp);

  /* now we can set our callback function */
  pcap_loop(handle, -1, sniff_logins, NULL);

  printf("\n* Exiting, sniffed %u logins\n", num_logins);

  /* cleanup */
  pcap_freecode(&fp);
  pcap_close(handle);

  exit(EXIT_SUCCESS);
}


/*
 * parse payload and search for SIP connections
 */
static void parse_payload(const conn_t *connection, 
			  unsigned char *payload, 
			  size_t payload_len)
{
  static sip_conn_t conn_table[MAX_SIP_CON];
  static int first=0, replace_entry=0;
  int i, ret, recorded=0;
  char buffer[SIP_LINE_LEN];
  char *payload_buffer=NULL;
  size_t payload_buffer_len=0;

  if(!first) {
    memset(conn_table, 0, sizeof(sip_conn_t) * MAX_SIP_CON);
    first=1;
  }
  
  if (payload_len <= 0)
    return;

  /* ignore packet if it contains binary data (for now */
  if(is_binary(payload, payload_len))
    return;

  /* Check if this is a recorded connection */
  for(i=0; i < MAX_SIP_CON; i++) {
    /* Known connection, check data */
    if(conn_table[i].active &&
       connection->server_ip == conn_table[i].server_ip &&
       connection->client_ip == conn_table[i].client_ip && 
       connection->server_port == conn_table[i].server_port &&
       connection->client_port == conn_table[i].client_port) 
    {
      struct in_addr cli, srv;
      cli.s_addr = connection->client_ip;
      srv.s_addr = connection->server_ip;
      printf("* New traffic on monitored connection %d (%s -> ", i, inet_ntoa(cli));
      printf("%s)\n", inet_ntoa(srv));
      recorded=1;

      /* No data recorded, just process normal */
      if(!strlen(conn_table[i].buffer)) {
	ret = parse_sip_proto(buffer, sizeof(buffer), payload, payload_len);
      }
      else {
        /* Append data to buffer and check string*/
	payload_buffer_len = payload_len + strlen(conn_table[i].buffer) + 1;
	payload_buffer = (char *)Malloc(payload_buffer_len);
	strncpy(payload_buffer, conn_table[i].buffer, payload_buffer_len);
	strncat(payload_buffer, (char *)payload, payload_buffer_len - strlen(payload_buffer));
        ret = parse_sip_proto(buffer, sizeof(buffer), (unsigned char *)payload_buffer, payload_buffer_len);
	free(payload_buffer);
      }
      /* error, ignore  */
      if(ret < 0) {
	return;
      }
      /* Found challenge response */
      if(ret) {
	struct in_addr cli, srv;
	cli.s_addr = connection->client_ip;
	srv.s_addr = connection->server_ip;
	printf("* Found challenge response (%s:%d <-> ", inet_ntoa(cli), connection->client_port);
	printf("%s:%d)\n", inet_ntoa(srv), connection->server_port);
	parse_n_write_login_data(conn_table[i].server_ip, conn_table[i].client_ip, buffer);
	conn_table[i].active=0;
      }
      /* copy new remaining data to struct */
      else if(!ret) {
	if(buffer[0] != '\0')
	  strncpy(conn_table[i].buffer, buffer, sizeof(conn_table[i].buffer) - 1);
      }

      break;
    }
  }

  /* Unknown connection, search for SIP */
  if(!recorded) {

    /* Check for SIP strings */
    if(!find_sip_client_proto(payload, payload_len))
      return; 

    ret = parse_sip_proto(buffer, sizeof(buffer), payload, payload_len);
    /* error, ignore  */
    if(ret < 0) {
      return;
    }
    /* Found challenge response */
    if(ret) {
      struct in_addr cli, srv;
      cli.s_addr = connection->client_ip;
      srv.s_addr = connection->server_ip;
      printf("* Found challenge response (%s:%d <-> ", inet_ntoa(cli), connection->client_port);
      printf("%s:%d)\n", inet_ntoa(srv), connection->server_port);
      parse_n_write_login_data(connection->server_ip, connection->client_ip, buffer);
    }
    /* Add to connection table for further checks */
    else if(!ret) {
      struct in_addr cli, srv;
      cli.s_addr = connection->client_ip;
      srv.s_addr = connection->server_ip;
      printf("* Adding %s:%d <-> ", inet_ntoa(cli), connection->client_port);
      printf("%s:%d to monitor list...", inet_ntoa(srv), connection->server_port);
      fflush(stdout);

      /* add to connection table */
      for(i=0; i < MAX_SIP_CON; i++) {
	if(!conn_table[i].active) {
	  recorded=1;
	  conn_table[i].active      = 1;
	  conn_table[i].client_ip   = connection->client_ip;
          conn_table[i].server_ip   = connection->server_ip;
	  conn_table[i].client_port = connection->client_port;
          conn_table[i].server_port = connection->server_port;
	  if(buffer[0] != '\0')
	    strncpy(conn_table[i].buffer, buffer, SIP_LINE_LEN);
	  printf("id %d\n", i);
	  break;
	}
      }
      /* if conntable full replace entries */
      if(!recorded) {
	struct in_addr cli, srv;
	cli.s_addr = conn_table[replace_entry].client_ip;
	srv.s_addr = conn_table[replace_entry].server_ip;
	printf("* Connection table full, replacing entry %d (%s -> ", replace_entry, inet_ntoa(cli));
	printf("%s)\n", inet_ntoa(srv));
	
	conn_table[replace_entry].active      = 1;
	conn_table[replace_entry].client_ip   = connection->client_ip;
	conn_table[replace_entry].server_ip   = connection->server_ip;
	conn_table[replace_entry].client_port = connection->client_port;
	conn_table[replace_entry].server_port = connection->server_port;
	if(buffer[0] != '\0')
	  strncpy(conn_table[replace_entry].buffer, buffer, SIP_LINE_LEN);
	
	if(replace_entry == MAX_SIP_CON - 1) replace_entry = 0;
	else replace_entry++;
      }
    }
  } /* if(!recorded) */

  
  return;
}

/*
 * check received packets
 */
static void sniff_logins(unsigned char *args, const struct pcap_pkthdr *header, const unsigned char *packet)
{
  static int count = 1;
  const struct ethernet_header *ethernet;
  const struct ip_header *ip;
  const struct tcp_header *tcp;
  const struct udp_header *udp;
  unsigned char *payload;
  conn_t connection;
  size_t size_ip=0, size_proto=0, size_payload=0;

  count++;

  ethernet = (struct ethernet_header *)(packet);


  ip = (struct ip_header *)(packet + SIZE_ETHERNET);
  size_ip = IP_HL(ip)*4;
  if (size_ip < 20) {
    printf("* Got packet with invalid IP header length (%d bytes), ignoring...\n", size_ip);
    return;
  }

  /* Assign IP's */
  connection.server_ip = ip->ip_src.s_addr;
  connection.client_ip = ip->ip_dst.s_addr;

  switch(ip->ip_p) {
  case IPPROTO_TCP:
    tcp = (struct tcp_header *)(packet + SIZE_ETHERNET + size_ip);
    size_proto = TH_OFF(tcp)*4;
    if (size_proto < 20) {
      printf("* Got packet with invalid TCP header length (%d bytes), ignoring...\n", size_proto);
      return;
    }
    connection.server_port = tcp->th_sport;
    connection.client_port = tcp->th_dport;
    break;
  case IPPROTO_UDP:
    udp = (struct udp_header *)(packet + SIZE_ETHERNET + size_ip);
    size_proto = sizeof(struct udp_header);
    connection.server_port = udp->uh_sport;
    connection.client_port = udp->uh_dport;
    break;
  default: 
    return;
  }

  payload = (unsigned char *)(packet + SIZE_ETHERNET + size_ip + size_proto);
  size_payload = ntohs(ip->ip_len) - (size_ip + size_proto);
  payload[size_payload] = '\0';

  if (size_payload > 0) {
    parse_payload(&connection, payload, size_payload);
  }
  
  return;
}

static int parse_sip_proto(char *out, 
			   size_t out_len, 
			   unsigned char *buffer, 
			   size_t buffer_len)
{
  char **lines;
  int num_lines, i, found=0, error=0;

  memset(out, 0, out_len);

  /* iterate through sip data */
  lines = stringtoarray((char *)buffer, '\n', &num_lines);

  for(i = 0 ; i < num_lines - 1 ; i++) {
    if(!strncmp(lines[i], "Authorization:", 14) && !found && !error) {
      if(out_len - 1 < strlen(lines[i])) {
	printf("* Buffer too small for line, ignoring...\n");
	error=1;
      }
      found = 1;
      strncpy(out, lines[i], out_len - 1);
    }
    free(lines[i]);
  }

  if(error) {
    free(lines[num_lines - 1]);
    return -1;
  }

  /* Challenge response sniffed */
  if(found) {
    free(lines[num_lines - 1]);
    return 1;  
  }

  /* check if last line had no newline (if challenge got split
   * between more packages 
   * If contains authorization line add to buffer for later adding
   */
  if(strlen(lines[num_lines-1])) {
    if(!strncmp(lines[num_lines-1], 
		"Authorization:", 
		strlen(lines[num_lines-1]) > 14 ? 14 : strlen(lines[num_lines-1]))) {
      if(out_len - 1 < strlen(lines[num_lines-1])) {
	printf("* Buffer too small for line, ignoring...\n");
	free(lines[num_lines-1]);
	return -1;
      }
      strncpy(out, lines[num_lines-1], out_len - 1);
    }
  }

  free(lines[num_lines-1]);

  return 0;
}

/* look in payload buffer for SIP protocol */
static int find_sip_client_proto(unsigned char *buffer, size_t len)
{
  int i;
  char c=0;

  /* If no register request, ignore atm */
  if(strncmp((char *)buffer, "REGISTER ", 8))
    return 0;
    
  for(i = 0; i < len; i++) {
    if(buffer[i] == 0x0a) { buffer[i] = 0x00; c = 0x0a; break; }
    else if(buffer[i] == 0x0d) { buffer[i] = 0x00; c = 0x0d; break; }
  }
  if(strstr((char *)buffer, " sip:") && strstr((char *)buffer, " SIP/")) {
    buffer[i] = c;
    return 1;
  }

  return 0;
}

static void manual_mode() 
{
  login_t login;
  memset(&login, 0, sizeof(login));

  /* Get input */
  printf("* Enter login information manually:\n\n");
  get_string_input(login.server,    sizeof(login.server),    "* Enter server IP: ");
  get_string_input(login.client,    sizeof(login.client),    "* Enter client IP: ");
  get_string_input(login.user,      sizeof(login.user),      "* Enter username : ");
  get_string_input(login.realm,     sizeof(login.realm),     "* Enter realm    : ");
  strncpy(login.method, "REGISTER", sizeof(login.method));
  get_string_input(login.uri,       sizeof(login.uri),       "* Enter URI      : ");
  get_string_input(login.nonce,     sizeof(login.nonce),     "* Enter nonce    : ");
  get_string_input(login.algorithm, sizeof(login.algorithm), "* Enter algoritm : ");
  get_string_input(login.hash,      sizeof(login.hash),      "* Enter hash     : ");

  Toupper(login.algorithm, strlen(login.algorithm));

  /* Write to file */
  write_login_data(&login, logins_file);
}

/*
 * Wrapper
 */

/* show usage */
static void usage(const char *err_msg)
{
  printf("Usage: SIPdump [ OPTIONS ] -d <dump file>\n\n"    \
         "       -d dump file  = file where captured logins will be written\n" \
         "       Options:                                                  \n" \
	 "       -m             = enter login data manually                \n" \
         "       -i <interface> = interface to listen on                   \n" \
	 "       -f <file>      = use tcpdump captured data file           \n" \
         "\n* %s\n", err_msg);
  exit(EXIT_FAILURE);
}

static void parse_n_write_login_data(uint32_t server, uint32_t client, char *buffer)
{
  login_t login_data;
  struct in_addr cli, srv;
 
  memset(&login_data, 0, sizeof(login_data));

  cli.s_addr = client;
  srv.s_addr = server;
  strncpy(login_data.server, inet_ntoa(srv), sizeof(login_data.server) - 1);
  strncpy(login_data.client, inet_ntoa(cli), sizeof(login_data.client) - 1);

  /* will be fixed soon */
  strncpy(login_data.method, "REGISTER", sizeof(login_data.method) - 1);

  /* Extract options from buffer */
  if(find_value("username=",  buffer, login_data.user,      sizeof(login_data.user)) < 0      ||
     find_value("realm=",     buffer, login_data.realm,     sizeof(login_data.realm)) < 0     ||
     find_value("uri=",       buffer, login_data.uri,       sizeof(login_data.uri)) < 0       ||
     find_value("nonce=",     buffer, login_data.nonce,     sizeof(login_data.nonce)) < 0     ||
     find_value("response=",  buffer, login_data.hash,      sizeof(login_data.hash)) < 0      )
    {
      printf("* Couldn't parse buffer (ignoring data):\n---------\n%s\n---------\n", buffer);
      return;
    }

  /* use MD5 is no algorithm given */
  if(find_value("algorithm=", buffer, login_data.algorithm, sizeof(login_data.algorithm)) < 0) {
      strncpy(login_data.algorithm, "MD5", sizeof(login_data.algorithm));
  }
  else {
    Toupper(login_data.algorithm, strlen(login_data.algorithm));
  }

  write_login_data(&login_data, logins_file);

  printf("* Wrote sniffed login %s -> %s (User: '%s') to dump file\n", 
	 login_data.client, login_data.server, login_data.user);

  num_logins++;
}

